"""add domain, page url unique keys

Revision ID: 178c6db7516
Revises: 25abb99fc57
Create Date: 2013-11-14 16:53:36.719799

"""

# revision identifiers, used by Alembic.
revision = '178c6db7516'
down_revision = '25abb99fc57'

import sqlalchemy as sa
from alembic import op


def upgrade():
    op.add_column(
        'pages',
        sa.Column('url_hash', type_=sa.String(128))
    )

    op.add_column(
        'domains',
        sa.Column('url_hash', type_=sa.String(128))
    )

    op.alter_column('domains', 'url', type_=sa.Text)
    op.alter_column('pages', 'url', type_=sa.Text)
    op.create_unique_constraint("uk_domain_url", "domains", ["url_hash"])
    op.create_unique_constraint("uk_page_url", "pages", ["url_hash"])


def downgrade():
    op.drop_constraint('uk_domain_url', 'domains', type_='unique')
    op.drop_constraint('uk_page_url', 'pages', type_='unique')
    op.alter_column('pages', 'url', type_=sa.String(2000))
    op.alter_column('domains', 'url', type_=sa.String(2000))
    op.drop_column('pages', 'url_hash')
    op.drop_column('domains', 'url_hash')



#!/usr/bin/python
# -*- coding: utf-8 -*-

from holmes.facters import Facter


class BodyFacter(Facter):
    @classmethod
    def get_fact_definitions(cls):
        return {}

    def get_facts(self):

        body = self.reviewer.current_html.cssselect('body')

        if not body:
            return

        self.review.data['page.body'] = body



#!/usr/bin/python
# -*- coding: utf-8 -*-

from holmes.validators.base import Validator


class OpenGraphValidator(Validator):

    @classmethod
    def get_open_graph_message(cls, value):
        return 'Some tags are missing: %s' % (', '.join(value))

    @classmethod
    def get_violation_definitions(cls):
        return {
            'absent.metatags.open_graph': {
                'title': 'Open Graph tags not found',
                'description': cls.get_open_graph_message
            },
        }

    def validate(self):
        meta_tags = self.get_meta_tags()

        if not meta_tags:
            return

        og = ['og:title', 'og:type', 'og:image', 'og:url']

        for item in meta_tags:
            og = self._missing_tags(item, 'og:title', og)
            og = self._missing_tags(item, 'og:type', og)
            og = self._missing_tags(item, 'og:image', og)
            og = self._missing_tags(item, 'og:url', og)

        if og:
            self.add_violation(
                key='absent.metatags.open_graph',
                value=og,
                points=50 * len(og)
            )

    def _missing_tags(self, item, key, current_tags):
        if not 'key' in item or not 'content' in item:
            return current_tags

        if item['key'] == key and item['content']:
            if key in current_tags:
                current_tags.remove(key)

        return current_tags

    def get_meta_tags(self):
        return self.review.data.get('meta.tags', None)



#!/usr/bin/python
# -*- coding: utf-8 -*-

from derpconf.config import Config  # NOQA


Config.define('WORKER_SLEEP_TIME', 10, 'Main loop sleep time', 'Worker')
Config.define('ZOMBIE_WORKER_TIME', 200,
              'Time to remove a Worker from API List (must be greater than WORKER_SLEEP_TIME + Validation time)', 'API')

Config.define('CONNECT_TIMEOUT_IN_SECONDS', 20, 'Number of seconds a connection can take.', 'Worker')
Config.define('REQUEST_TIMEOUT_IN_SECONDS', 60, 'Number of seconds a request can take.', 'Worker')

Config.define('HOLMES_API_URL', 'http://localhost:2368', 'URL that Worker will communicate with API', 'Worker')

Config.define('LOG_LEVEL', 'ERROR', 'Default log level', 'Logging')
Config.define('LOG_FORMAT', '%(asctime)s:%(levelname)s %(module)s - %(message)s',
              'Log Format to be used when writing log messages', 'Logging')
Config.define('LOG_DATE_FORMAT', '%Y-%m-%d %H:%M:%S',
              'Date Format to be used when writing log messages.', 'Logging')

Config.define('FACTERS', [], 'List of classes to get facts about a website', 'Review')
Config.define('VALIDATORS', [], 'List of classes to validate a website', 'Review')
Config.define('REVIEW_EXPIRATION_IN_SECONDS', 6 * 60 * 60, 'Number of seconds that a review expires in.', 'Review')

Config.define('MAX_ENQUEUE_BUFFER_LENGTH', 1000,
              'Number of urls to enqueue before submitting to the /pages route', 'Validators')

# Reference data retrieved from HTTP Archive in 06-jan-2014
Config.define('MAX_IMG_REQUESTS_PER_PAGE', 40,
              'Maximum number of images per page', 'Image Request Validator')
Config.define('MAX_KB_SINGLE_IMAGE', 26,
              'Maximum size of a single image', 'Image Request Validator')
Config.define('MAX_IMG_KB_PER_PAGE', 1028,
              'Maximum size of images per page', 'Image Request Validator')

# Reference data retrieved from HTTP Archive in 06-jan-2014
Config.define('MAX_CSS_REQUESTS_PER_PAGE', 8,
              'Maximum number of external stylesheets per page', 'CSS Request Validator')
Config.define('MAX_CSS_KB_PER_PAGE_AFTER_GZIP', 46,
              'Maximum size of stylesheets per page after gzip', 'CSS Request Validator')

# Reference data retrieved from HTTP Archive in 06-jan-2014
Config.define('MAX_JS_REQUESTS_PER_PAGE', 17,
              'Maximum number of external scripts per page', 'JS Request Validator')
Config.define('MAX_JS_KB_PER_PAGE_AFTER_GZIP', 272,
              'Maximum size of scripts per page after gzip', 'JS Request Validator')

Config.define('MAX_TITLE_SIZE', 70,
              'Title tags longer than 70 characters may be truncated in the results',
              'Title Validator')

Config.define('ORIGIN', '*', 'Access Control Allow Origin header value', 'Web')

Config.define('HTTP_PROXY_HOST', None, 'HTTP Proxy Host to use', 'Web')
Config.define('HTTP_PROXY_PORT', None, 'HTTP Proxy Port to use', 'Web')

Config.define('API_PROXY_HOST', None, 'HTTP Proxy Host to use to connect to the API', 'Web')
Config.define('API_PROXY_PORT', None, 'HTTP Proxy Port to use to connect to the API', 'Web')

Config.define('COMMIT_ON_REQUEST_END', True, 'Commit on request end', 'DB')

Config.define('REDISHOST', 'localhost', 'Redis host', 'Redis')
Config.define('REDISPORT', 7575, 'Redis port', 'Redis')
Config.define('REDISPASS', None, 'Redis password in case of auth', 'Redis')

Config.define('REQUIRED_META_TAGS', [], 'List of required meta tags', 'Meta tag Validator')

Config.define('SCHEMA_ORG_ITEMTYPE', [], 'List of Schema.Org ItemType', 'Schema.Org ItemType')

Config.define('FORCE_CANONICAL', False, 'Force canonical', 'Force canonical')

Config.define('BLACKLIST_DOMAIN', [], 'Domain blacklist', 'Domain blacklist')

Config.define('ERROR_HANDLERS', [], 'List of classes to handle errors', 'General')

# SENTRY ERROR HANDLER
Config.define('USE_SENTRY', False, 'If set to true errors will be sent to sentry.', 'Sentry')
Config.define('SENTRY_DSN_URL', '', 'URL to use as sentry DSN.', 'Sentry')

HOUR = 60 * 60
Config.define('PAGE_COUNT_EXPIRATION_IN_SECONDS', HOUR, 'Expiration for the cache key for each domain page count', 'Cache')
Config.define('VIOLATION_COUNT_EXPIRATION_IN_SECONDS', HOUR, 'Expiration for the cache key for each domain violation count', 'Cache')
Config.define('ACTIVE_REVIEW_COUNT_EXPIRATION_IN_SECONDS', HOUR, 'Expiration for the cache key for each domain violation count', 'Cache')
Config.define('URL_LOCK_EXPIRATION_IN_SECONDS', 30, 'Expiration for the url lock for each url', 'Cache')



#!/usr/bin/python
# -*- coding: utf-8 -*-

from uuid import uuid4
from datetime import datetime

import sqlalchemy as sa

from holmes.models import Base


class Worker(Base):
    __tablename__ = "workers"

    id = sa.Column(sa.Integer, primary_key=True)
    uuid = sa.Column('uuid', sa.String(36), default=uuid4, nullable=False)
    last_ping = sa.Column('last_ping', sa.DateTime, default=datetime.now, nullable=False)

    current_url = sa.Column('current_url', sa.Text)

    def __str__(self):
        return 'Worker %s' % str(self.uuid)

    def __repr__(self):
        return str(self)

    @property
    def working(self):
        return self.current_url is not None

    def to_dict(self):
        return {
            'uuid': str(self.uuid),
            'last_ping': str(self.last_ping),
            'current_url': self.current_url,
            'working': self.working
        }

    @classmethod
    def by_uuid(cls, uuid, db):
        return db.query(Worker).filter(Worker.uuid == uuid).first()



"""Page table indexes for next job

Revision ID: 10e3ffb42fc9
Revises: 34bd7155169
Create Date: 2013-11-27 15:20:16.940698

"""

# revision identifiers, used by Alembic.
revision = '10e3ffb42fc9'
down_revision = '34bd7155169'

from alembic import op


def upgrade():
    op.create_index('idx_pages_last_review_date', 'pages', ['last_review_started_date'])
    op.create_index('idx_pages_created_date', 'pages', ['created_date'])


def downgrade():
    op.drop_index('idx_pages_last_review_date', 'pages')
    op.drop_index('idx_pages_created_date', 'pages')



#!/usr/bin/python
# -*- coding: utf-8 -*-

import logging

from cow.server import Server
from cow.plugins.sqlalchemy_plugin import SQLAlchemyPlugin
from cow.plugins.redis_plugin import RedisPlugin, CowRedisClient
from tornado.httpclient import AsyncHTTPClient
#from toredis import Client

from holmes.handlers.worker import WorkerHandler, WorkersHandler, WorkersInfoHandler
from holmes.handlers.worker_state import WorkerStateHandler
from holmes.handlers.page import (
    PageHandler, PagesHandler, PageReviewsHandler, PageViolationsPerDayHandler
)
from holmes.handlers.violation import MostCommonViolationsHandler
from holmes.handlers.review import (
    ReviewHandler, LastReviewsHandler
)
from holmes.handlers.next_job import NextJobHandler
from holmes.handlers.domains import (
    DomainsHandler, DomainDetailsHandler, DomainViolationsPerDayHandler,
    DomainReviewsHandler
)
from holmes.handlers.search import (
    SearchHandler
)
from holmes.handlers.bus import EventBusHandler
from holmes.event_bus import EventBus, NoOpEventBus
from holmes.utils import load_classes
from holmes.models import Key
from holmes.cache import Cache


def main():
    AsyncHTTPClient.configure("tornado.curl_httpclient.CurlAsyncHTTPClient")
    HolmesApiServer.run()


class HolmesApiServer(Server):
    def __init__(self, db=None, debug=None, *args, **kw):
        super(HolmesApiServer, self).__init__(*args, **kw)

        self.force_debug = debug
        self.db = db

    def initialize_app(self, *args, **kw):
        super(HolmesApiServer, self).initialize_app(*args, **kw)

        self.application.db = None

        if self.force_debug is not None:
            self.debug = self.force_debug

    def get_handlers(self):
        handlers = [
            (r'/most-common-violations/?', MostCommonViolationsHandler),
            (r'/last-reviews/?', LastReviewsHandler),
            (r'/workers/?', WorkersHandler),
            (r'/workers/info/?', WorkersInfoHandler),
            (r'/worker/([a-z0-9-]*)/(alive|dead)/?', WorkerHandler),
            (r'/worker/([a-z0-9-]*)/(start|complete)/?', WorkerStateHandler),
            (r'/page/([a-z0-9-]*)/review/([a-z0-9-]*)/?', ReviewHandler),
            (r'/page/([a-z0-9-]*)/reviews/?', PageReviewsHandler),
            (r'/page/([a-z0-9-]*)/violations-per-day/?', PageViolationsPerDayHandler),
            (r'/page/([a-z0-9-]*)/?', PageHandler),
            (r'/search/?', SearchHandler),
            (r'/page/?', PageHandler),
            (r'/pages/?', PagesHandler),
            (r'/domains/?', DomainsHandler),
            (r'/domains/([^/]+)/?', DomainDetailsHandler),
            (r'/domains/([^/]+)/violations-per-day/?', DomainViolationsPerDayHandler),
            (r'/domains/([^/]+)/reviews/?', DomainReviewsHandler),
            (r'/next/?', NextJobHandler),
            (r'/events/?', EventBusHandler),
        ]

        return tuple(handlers)

    def get_plugins(self):
        return [
            SQLAlchemyPlugin,
            RedisPlugin
        ]

    def after_start(self, io_loop):
        if self.db is not None:
            self.application.db = self.db
        else:
            self.application.db = self.application.get_sqlalchemy_session()

        if self.debug:
            import sqltap
            sqltap.start()

        self.application.facters = self._load_facters()
        self.application.validators = self._load_validators()
        self.application.error_handlers = [handler(self.application.config) for handler in self._load_error_handlers()]

        self.application.fact_definitions = {}
        self.application.violation_definitions = {}

        for facter in self.application.facters:
            self.application.fact_definitions.update(facter.get_fact_definitions())

        self._insert_keys(self.application.fact_definitions)

        for validator in self.application.validators:
            self.application.violation_definitions.update(validator.get_violation_definitions())

        self._insert_keys(self.application.violation_definitions)

        self.application.event_bus = NoOpEventBus(self.application)
        self.application.http_client = AsyncHTTPClient(io_loop=io_loop)
        self.connect_pub_sub(io_loop)

        self.application.cache = Cache(self.application)

    def _insert_keys(self, keys):
        for name in keys.keys():
            key = Key.get_or_create(self.application.db, name)
            keys[name]['key'] = key
            self.application.db.add(key)
            self.application.db.flush()
            self.application.db.commit()

    def _load_validators(self):
        return load_classes(default=self.config.VALIDATORS)

    def _load_facters(self):
        return load_classes(default=self.config.FACTERS)

    def _load_error_handlers(self):
        return load_classes(default=self.config.ERROR_HANDLERS)

    def before_end(self, io_loop):
        self.application.db.remove()

        if self.debug:
            import sqltap
            statistics = sqltap.collect()
            sqltap.report(statistics, "report.html")

    def connect_pub_sub(self, io_loop):
        host = self.application.config.get('REDISHOST')
        port = self.application.config.get('REDISPORT')

        logging.info("Connecting pubsub to redis at %s:%d" % (host, port))

        self.application.redis_pub_sub = CowRedisClient(io_loop=io_loop)
        self.application.redis_pub_sub.authenticated = False
        self.application.redis_pub_sub.connect(host, port, callback=self.has_connected(self.application))

    def has_connected(self, application):
        def handle(*args, **kw):
            password = application.config.get('REDISPASS', None)
            if password:
                application.redis_pub_sub.auth(password, callback=self.handle_authenticated(application))
            else:
                self.handle_authenticated(application)()
        return handle

    def handle_authenticated(self, application):
        def handle(*args, **kw):
            application.redis_pub_sub.authenticated = True
            application.event_bus = EventBus(self.application)  # can now connect to redis using pubsub

        return handle

if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

from uuid import uuid4
from datetime import datetime

import sqlalchemy as sa
from sqlalchemy.orm import relationship

from holmes.models import Base


class Page(Base):
    __tablename__ = "pages"

    id = sa.Column(sa.Integer, primary_key=True)
    url = sa.Column('url', sa.String(2000), nullable=False)
    url_hash = sa.Column('url_hash', sa.String(128), nullable=False)
    uuid = sa.Column('uuid', sa.String(36), default=uuid4, nullable=False)
    created_date = sa.Column('created_date', sa.DateTime, default=datetime.utcnow, nullable=False)

    domain_id = sa.Column('domain_id', sa.Integer, sa.ForeignKey('domains.id'))

    reviews = relationship("Review", backref="page", foreign_keys='[Review.page_id]')

    last_review_id = sa.Column('last_review_id', sa.Integer, sa.ForeignKey('reviews.id'))
    last_review = relationship("Review", foreign_keys=[last_review_id])
    last_review_date = sa.Column('last_review_date', sa.DateTime, nullable=True)
    last_review_uuid = sa.Column('last_review_uuid', sa.String(36), nullable=True)

    last_modified = sa.Column('last_modified', sa.DateTime, nullable=True)
    expires = sa.Column('expires', sa.DateTime, nullable=True)

    violations_count = sa.Column('violations_count', sa.Integer, server_default='0', nullable=False)

    def to_dict(self):
        return {
            'uuid': str(self.uuid),
            'url': self.url,
            'lastModified': self.last_modified,
            'expires': self.expires,
        }

    def __str__(self):
        return str(self.uuid)

    def __repr__(self):
        return str(self)

    def get_violations_per_day(self, db):
        from holmes.models import Review, Violation  # Prevent circular dependency

        violations = db \
            .query(
                sa.func.year(Review.completed_date).label('year'),
                sa.func.month(Review.completed_date).label('month'),
                sa.func.day(Review.completed_date).label('day'),
                sa.func.count(Violation.id).label('violation_count'),
                sa.func.sum(Violation.points).label('violation_points')
            ).join(
                Page, Page.id == Review.page_id
            ).join(
                Violation, Violation.review_id == Review.id
            ).filter(Review.is_complete == True).filter(Review.page_id == self.id) \
            .group_by(
                sa.func.year(Review.completed_date),
                sa.func.month(Review.completed_date),
                sa.func.day(Review.completed_date),
            ) \
            .order_by(Review.completed_date) \
            .all()

        result = []

        for day in violations:
            dt = "%d-%d-%d" % (day.year, day.month, day.day)
            result.append({
                "completedAt": dt,
                "violation_count": int(day.violation_count),
                "violation_points": int(day.violation_points)
            })

        return result

    @classmethod
    def by_uuid(cls, uuid, db):
        return db.query(Page).filter(Page.uuid == uuid).first()



"""create page table

Revision ID: 1fa4469fb63b
Revises: 34a477a7f285
Create Date: 2013-11-13 10:21:17.150033

"""

# revision identifiers, used by Alembic.
revision = '1fa4469fb63b'
down_revision = '34a477a7f285'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'pages',
        sa.Column('id', sa.Integer, primary_key=True),
        sa.Column('url', sa.String(2000), nullable=False),
        sa.Column('uuid', sa.String(36), nullable=False),
        sa.Column('created_date', sa.DateTime, nullable=False),
        sa.Column('last_review_started_date', sa.DateTime, nullable=True),
        sa.Column('domain_id', sa.Integer)
    )

    op.create_foreign_key(
        "fk_page_domain", "pages",
        "domains", ["domain_id"], ["id"]
    )


def downgrade():
    op.drop_constraint('fk_page_domain', 'pages', type_="foreignkey")
    op.drop_table('pages')



#!/usr/bin/python
# -*- coding: utf-8 -*-

import sqlalchemy as sa

from holmes.models import Base, JsonType
from holmes.models.keys import Key


class Violation(Base):
    __tablename__ = "violations"

    id = sa.Column(sa.Integer, primary_key=True)
    # key = sa.Column('key', sa.String(2000), nullable=False)
    value = sa.Column('value', JsonType, nullable=True)
    #title = sa.Column('title', sa.String(2000), nullable=False)
    #description = sa.Column('description', sa.Text, nullable=False)
    points = sa.Column('points', sa.Integer, nullable=False)

    review_id = sa.Column('review_id', sa.Integer, sa.ForeignKey('reviews.id'))
    # review comes from Review relationship
    key_id = sa.Column('key_id', sa.Integer, sa.ForeignKey('keys.id'))

    def __str__(self):
        return '%s: %s' % (self.key.name, self.value)

    def __repr__(self):
        return str(self)

    def to_dict(self, violation_definitions):
        definition = violation_definitions.get(self.key.name, {})

        return {
            'key': self.key.name,
            'title': definition.get('title', 'undefined'),
            'description': definition.get('description', lambda value: value)(self.value),
            'points': self.points
        }

    @classmethod
    def get_most_common_violations(cls, db, violation_definitions, limit=5):
        violations = []

        results = db.query(Violation.key_id, Key.name, sa.func.count(Violation.id).label('count')) \
                    .outerjoin(Key) \
                    .group_by(Violation.key_id) \
                    .order_by('count desc') \
                    .limit(limit)

        for key_id, key_name, count in results:
            definition = violation_definitions.get(key_name, {})
            violations.append({
                "key": key_name,
                "title": definition.get('title', 'undefined'),
                "count": count
            })

        return violations



#!/usr/bin/python
# -*- coding: utf-8 -*-

from holmes.validators.base import Validator


class SchemaOrgItemTypeValidator(Validator):
    @classmethod
    def get_itemscope_message(cls, value=None):
        return 'In order to conform to schema.org definition ' \
               'of a webpage, the body tag must feature an ' \
               'itemscope attribute.'

    @classmethod
    def get_itemtype_message(cls, value=None):
        return 'In order to conform to schema.org definition ' \
               'of a webpage, the body tag must feature an ' \
               'itemtype attribute.'

    @classmethod
    def get_invalid_itemtype_message(cls, value=None):
        url = 'http://schema.org/WebPage'
        return 'In order to conform to schema.org definition ' \
               'of a webpage, the body tag must feature an ' \
               'itemtype attribute with a value of "%s" or ' \
               'one of its more specific types.' % url

    @classmethod
    def get_violation_definitions(cls):
        return {
            'absent.schema.itemscope': {
                'title': 'itemscope attribute missing in body',
                'description': cls.get_itemscope_message
            },
            'absent.schema.itemtype': {
                'title': 'itemtype attribute missing in body',
                'description': cls.get_itemtype_message
            },
            'invalid.schema.itemtype': {
                'title': 'itemtype attribute is invalid',
                'description': cls.get_invalid_itemtype_message
            }
        }

    def validate(self):
        body = self.get_body()

        if not body:
            return

        if len(body) == 1:

            body = body[0]
            attributes = body.keys()

            if 'itemscope' not in attributes:
                self.add_violation(
                    key='absent.schema.itemscope',
                    value=None,
                    points=10)

            has_itemtype = 'itemtype' in attributes
            if not has_itemtype:
                self.add_violation(
                    key='absent.schema.itemtype',
                    value=None,
                    points=10
                )

            itemtype_value = self.reviewer.config.SCHEMA_ORG_ITEMTYPE
            if has_itemtype and body.get('itemtype') not in itemtype_value:
                self.add_violation(
                    key='invalid.schema.itemtype',
                    value=None,
                    points=10
                )

    def get_body(self):
        return self.review.data.get('page.body', None)



#!/usr/bin/python
# -*- coding: utf-8 -*-

from holmes.validators.base import Validator


class ImageWithoutAltAttributeValidator(Validator):
    @classmethod
    def get_empy_anchors_message(cls, value):
        result = []
        for src, name in value:
            data = '<a href="%s" target="_blank">%s</a>' % (src, name)
            result.append(data)

        return 'Images without alt text are not good for ' \
               'Search Engines. Images without alt were ' \
               'found for: %s.' % (', '.join(result))

    @classmethod
    def get_violation_definitions(cls):
        return {
            'invalid.images.alt': {
                'title': 'Image(s) without alt attribute',
                'description': cls.get_empy_anchors_message
            }
        }

    def validate(self):
        imgs = self.get_imgs()

        result = []
        for img in imgs:
            src = img.get('src')
            if not src:
                continue

            if not self.is_valid(src):
                continue

            is_absolute = self.is_absolute(src)

            if not is_absolute:
                src = self.rebase(src)

            if not img.get('alt'):
                name = src.rsplit('/', 1)[-1]
                result.append((src, name))

        if result:
            self.add_violation(
                key='invalid.images.alt',
                value=result,
                points=20 * len(result)
            )

    def get_imgs(self):
        return self.review.data.get('page.all_images', None)



#!/usr/bin/python
# -*- coding: utf-8 -*-

from tornado.concurrent import return_future

from holmes.models import Domain


class Cache(object):
    def __init__(self, application):
        self.application = application
        self.redis = self.application.redis
        self.db = self.application.db
        self.config = self.application.config

    def get_domain_name(self, domain_name):
        if isinstance(domain_name, Domain):
            return domain_name.name

        return domain_name

    @return_future
    def has_key(self, key, callback):
        self.redis.exists(key, callback)

    @return_future
    def increment_violations_count(self, domain_name, increment=1, callback=None):
        self.increment_count(
            'violation-count',
            domain_name,
            lambda domain: domain.get_violation_data(self.db),
            increment,
            callback
        )

    @return_future
    def increment_active_review_count(self, domain_name, increment=1, callback=None):
        self.increment_count(
            'active-review-count',
            domain_name,
            lambda domain: domain.get_active_review_count(self.db),
            increment,
            callback
        )

    @return_future
    def increment_page_count(self, domain_name, increment=1, callback=None):
        self.increment_count(
            'page-count',
            domain_name,
            lambda domain: domain.get_page_count(self.db),
            increment,
            callback
        )

    def increment_count(self, key, domain_name, get_default_method, increment=1, callback=None):
        key = '%s-%s' % (self.get_domain_name(domain_name), key)
        self.has_key(key, self.handle_has_key(key, domain_name, get_default_method, increment, callback))

    def handle_has_key(self, key, domain_name, get_default_method, increment, callback):
        def handle(has_key):
            domain = domain_name
            if domain and not isinstance(domain, Domain):
                domain = Domain.get_domain_by_name(domain_name, self.db)

            if not domain:
                callback(None)
                return

            if has_key:
                self.redis.incrby(key, increment, callback=callback)
            else:
                value = get_default_method(domain)
                self.redis.set(key, value + increment, callback=callback)

        return handle

    @return_future
    def get_page_count(self, domain_name, callback=None):
        self.get_count(
            'page-count',
            domain_name,
            int(self.config.PAGE_COUNT_EXPIRATION_IN_SECONDS),
            lambda domain: domain.get_page_count(self.db),
            callback=callback
        )

    @return_future
    def get_violation_count(self, domain_name, callback=None):
        self.get_count(
            'violation-count',
            domain_name,
            int(self.config.PAGE_COUNT_EXPIRATION_IN_SECONDS),
            lambda domain: domain.get_violation_data(self.db),
            callback=callback
        )

    @return_future
    def get_active_review_count(self, domain_name, callback=None):
        self.get_count(
            'active-review-count',
            domain_name,
            int(self.config.ACTIVE_REVIEW_COUNT_EXPIRATION_IN_SECONDS),
            lambda domain: domain.get_active_review_count(self.db),
            callback=callback
        )

    def get_count(self, key, domain_name, expiration, get_count_method, callback=None):
        cache_key = '%s-%s' % (self.get_domain_name(domain_name), key)
        self.redis.get(cache_key, callback=self.handle_get_count(key, domain_name, expiration, get_count_method, callback))

    def handle_get_count(self, key, domain_name, expiration, get_count_method, callback):
        def handle(count):
            if count is not None:
                callback(int(count))
                return

            domain = domain_name
            if domain and not isinstance(domain, Domain):
                domain = Domain.get_domain_by_name(domain_name, self.db)

            if not domain:
                callback(None)
                return

            count = get_count_method(domain)

            cache_key = '%s-%s' % (domain.name, key)

            self.redis.setex(
                key=cache_key,
                value=int(count),
                seconds=expiration,
                callback=self.handle_set_count(count, callback)
            )

        return handle

    def handle_set_count(self, count, callback):
        def handle(*args, **kw):
            callback(count)

        return handle

    @return_future
    def lock_page(self, url, callback=None):
        expiration = self.config.URL_LOCK_EXPIRATION_IN_SECONDS

        self.redis.setex(
            key='%s-lock' % url,
            value=1,
            seconds=expiration,
            callback=callback
        )

    @return_future
    def has_lock(self, url, callback=None):
        self.redis.get(
            key='%s-lock' % url,
            callback=self.handle_get_lock_page(url, callback)
        )

    def handle_get_lock_page(self, url, callback):
        def handle(value):
            callback(value == '1')

        return handle



#!/usr/bin/python
# -*- coding: utf-8 -*-

from tornado import gen

from holmes.handlers import BaseHandler
from holmes.models import Violation


class MostCommonViolationsHandler(BaseHandler):

    @gen.coroutine
    def get(self):
        violations = Violation.get_most_common_violations(self.db, self.application.violation_definitions)

        result = []
        for item in violations:
            result.append({'name': item['title'], 'count': item['count']})

        self.write_json(result)
        self.finish()



#!/usr/bin/python
# -*- coding: utf-8 -*-

import logging
from collections import defaultdict


class NoOpEventBus(object):
    def __init__(self, application):
        self.application = application

    def subscribe(self, channel, uuid, handler):
        pass

    def unsubscribe(self, channel, uuid):
        pass

    def publish(self, message):
        pass

    def flush(self):
        pass


class EventBus(object):
    def __init__(self, application):
        self.application = application
        self.handlers = defaultdict(dict)

        self.publish_items = []
        self.application.redis_pub_sub.subscribe('events', self.on_message)

    def subscribe(self, channel, uuid, handler):
        self.handlers[channel][uuid] = handler

    def unsubscribe(self, channel, uuid):
        if self.handlers.get(channel, {}).get(uuid, None) is None:
            return

        del self.handlers[channel][uuid]

    def publish(self, message):
        self.publish_items.append(('events', message))

    def flush(self):
        for channel, message in self.publish_items:
            logging.debug('Publishing message to %s...' % channel)
            self.application.redis.publish(channel, message)
        self.publish_items = []

    def on_message(self, message):
        if message is None:
            return

        msg_type, msg_channel, msg_value = message

        logging.debug("%s on %s for %s" % (msg_type, msg_channel, msg_value))

        if msg_type != 'message':
            return

        for handler in self.handlers.get(msg_channel, {}).values():
            handler(msg_value)



#!/usr/bin/python
# -*- coding: utf-8 -*-

from datetime import datetime, timedelta

from ujson import dumps

import sqlalchemy as sa

from holmes.models.worker import Worker
from holmes.handlers import BaseHandler


class WorkerHandler(BaseHandler):

    def post(self, worker_uuid, i_am='alive'):
        self._remove_zombie_workers()

        worker = Worker.by_uuid(worker_uuid, self.db)

        if worker:
            if i_am == 'alive':
                worker.last_ping = datetime.now()
            elif i_am == 'dead':
                self.db.delete(worker)
        else:
            worker = Worker(uuid=worker_uuid)
            self.db.add(worker)

        self.db.flush()

        self.application.event_bus.publish(dumps({
            'type': 'worker-status',
            'workerId': str(worker.uuid)
        }))

        self.write(str(worker_uuid))

    def _remove_zombie_workers(self):
        self.db.flush()

        dt = datetime.now() - timedelta(seconds=self.application.config.ZOMBIE_WORKER_TIME)

        workers_to_delete = self.db.query(Worker).filter(Worker.last_ping < dt).order_by(Worker.id).all()

        for worker in workers_to_delete:
            self.db.delete(worker)

        self.db.flush()


class WorkersHandler(BaseHandler):

    def get(self):
        workers = self.db.query(Worker).all()

        workers_json = [worker.to_dict() for worker in workers]
        self.write_json(workers_json)


class WorkersInfoHandler(BaseHandler):

    def get(self):
        total_workers = self.db \
            .query(sa.func.count(Worker.id).label('count')) \
            .one()

        inactive_workers = self.db \
            .query(sa.func.count(Worker.id).label('count')) \
            .filter(Worker.current_url == None) \
            .one()

        workers_info = {
            'total': total_workers.count,
            'inactive': inactive_workers.count,
            'active': total_workers.count - inactive_workers.count,
        }
        self.write_json(workers_info)



#!/usr/bin/python
# -*- coding: utf-8 -*-

import urlparse


class Validator(object):
    def __init__(self, reviewer):
        self.reviewer = reviewer
        self.url_buffer = set()

    @classmethod
    def get_violation_definitions(cls):
        raise NotImplementedError

    def is_absolute(self, url):
        return bool(urlparse.urlparse(url).scheme)

    def rebase(self, url):
        return urlparse.urljoin(self.page_url.rstrip('/'), url)

    @property
    def page_uuid(self):
        return self.reviewer.page_uuid

    @property
    def page_url(self):
        return self.reviewer.page_url

    @property
    def review(self):
        return self.reviewer.review_dao

    @property
    def config(self):
        return self.reviewer.config

    def validate(self):
        return True

    def to_gzip(self, content):
        return content.encode('zip')

    def enqueue(self, url):
        self.reviewer.enqueue(url)

    def add_fact(self, key, value):
        self.reviewer.add_fact(key, value)

    def add_violation(self, key, value, points):
        self.reviewer.add_violation(key, value, points)

    def is_valid(self, url):
        try:
            return urlparse.urlparse(url)
        except ValueError:
            return None

    def test_url(self, url, response, broken_link_callback=None, moved_link_callback=None):
        status = response.status_code

        if status > 399:
            if broken_link_callback:
                broken_link_callback(url, response)
            return False

        if status == 302 or status == 307:
            if moved_link_callback:
                moved_link_callback(url, response)
            return False

        if response.url.rstrip('/') != url.rstrip('/'):
            return False

        return True

    def send_url(self, url, response):
        if self.test_url(url, response, self.broken_link_violation, self.moved_link_violation):
            self.url_buffer.add(url)

        if len(self.url_buffer) > self.config.MAX_ENQUEUE_BUFFER_LENGTH:
            self.flush()

    def flush(self):
        if not self.url_buffer:
            return

        self.enqueue(self.url_buffer)
        self.url_buffer = set()

    def broken_link_violation(self):
        raise NotImplementedError('broken_link_violation method need to be implemented by {0}'.format(self.__class__.__name__))

    def moved_link_violation(self):
        raise NotImplementedError('moved_link_violation method need to be implemented by {0}'.format(self.__class__.__name__))



#!/usr/bin/python
# -*- coding: utf-8 -*-

from random import choice
from datetime import datetime, timedelta

from tornado import gen

from holmes.handlers import BaseHandler
from holmes.models.page import Page


class NextJobHandler(BaseHandler):

    @gen.coroutine
    def get(self):
        dt = datetime.now() - timedelta(seconds=self.application.config.REVIEW_EXPIRATION_IN_SECONDS)

        pages_in_need_of_review = self.db.query(Page.uuid, Page.url) \
            .filter(Page.last_review == None)[:50]

        if len(pages_in_need_of_review) == 0:
            pages_in_need_of_review = self.db.query(Page.uuid, Page.url) \
                .filter(
                    Page.last_review != None,
                    Page.last_review_date < dt
                ) \
                .order_by(Page.created_date)[:50]

            if len(pages_in_need_of_review) == 0:
                self.write('')
                self.finish()
                return

        page = choice(pages_in_need_of_review)

        self.write_json({
            'page': str(page.uuid),
            'url': page.url
        })



#!/usr/bin/python
# -*- coding: utf-8 -*-

import sqlalchemy.types as types
from sqlalchemy.ext.declarative import declarative_base
from ujson import dumps, loads
Base = declarative_base()


class JsonType(types.TypeDecorator):
    """Represents an immutable structure as a json-encoded string.

    Usage::

        JSONEncodedDict(255)

    """

    impl = types.Unicode

    def process_bind_param(self, value, dialect):
        if value is not None:
            value = dumps(value)

        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            value = loads(value)
        return value


from holmes.models.domain import Domain  # NOQA
from holmes.models.page import Page  # NOQA
from holmes.models.review import Review  # NOQA
from holmes.models.fact import Fact  # NOQA
from holmes.models.violation import Violation  # NOQA
from holmes.models.worker import Worker  # NOQA
from holmes.models.keys import Key  # NOQA



#!/usr/bin/python
# -*- coding: utf-8 -*-

from holmes.utils import get_domain_from_url
from holmes.validators.base import Validator


class BlackListValidator(Validator):
    @classmethod
    def get_blacklist_message(cls, value):
        return 'Some links are blacklisted: %s' % (
            ', '.join(
                    [
                        '<a href="%s" target="_blank">Link #%s</a>' % (url, index)
                        for index, url in enumerate(value)
                    ]
                )
            )

    @classmethod
    def get_violation_definitions(cls):
        return {
            'blacklist.domains': {
                'title': 'Domain Blacklist',
                'description': cls.get_blacklist_message
            }
        }

    def validate(self):
        domains = []

        links = self.get_links()

        for link in links:
            href = link.get('href')

            if not self.is_valid(href):
                continue

            link_domain, link_domain_url = get_domain_from_url(href)
            if link_domain in self.reviewer.config.BLACKLIST_DOMAIN:
                domains.append(href)

        if domains:
            self.add_violation(
                key='blacklist.domains',
                value=domains,
                points=100 * len(domains)
            )

    def get_links(self):
        return self.review.data.get('page.all_links', None)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import logging
from uuid import uuid4
from os.path import join

import requests
from ujson import dumps, loads
from requests.exceptions import ConnectionError
from sheep import Shepherd
from colorama import Fore, Style
from octopus import TornadoOctopus

from holmes import __version__
from holmes.config import Config
from holmes.reviewer import Reviewer, InvalidReviewError
from holmes.utils import load_classes


class HolmesWorker(Shepherd):
    def initialize(self):
        self.uuid = uuid4().hex
        self.working = True

        self.facters = self._load_facters()
        self.validators = self._load_validators()
        self.error_handlers = [handler(self.config) for handler in self._load_error_handlers()]

        logging.debug('Starting Octopus with %d concurrent threads.' % self.options.concurrency)
        self.otto = TornadoOctopus(
            concurrency=self.options.concurrency, cache=self.options.cache,
            connect_timeout_in_seconds=self.config.CONNECT_TIMEOUT_IN_SECONDS,
            request_timeout_in_seconds=self.config.REQUEST_TIMEOUT_IN_SECONDS
        )
        self.otto.start()

    def _load_error_handlers(self):
        return load_classes(default=self.config.ERROR_HANDLERS)

    def handle_error(self, exc_type, exc_value, tb):
        for handler in self.error_handlers:
            handler.handle_exception(
                exc_type, exc_value, tb, extra={
                    'worker-uuid': self.uuid,
                    'holmes-version': __version__
                }
            )

    def config_parser(self, parser):
        parser.add_argument(
            '--concurrency',
            '-t',
            type=int,
            default=10,
            help='Number of threads (or async http requests) to use for Octopus (doing GETs concurrently)'
        )

        parser.add_argument(
            '--cache', default=False, action='store_true',
            help='Whether http requests should be cached by Octopus.'
        )

    @property
    def proxies(self):
        proxies = None
        if self.config.HTTP_PROXY_HOST is not None:
            proxy = "%s:%s" % (self.config.HTTP_PROXY_HOST, self.config.HTTP_PROXY_PORT)
            http_proxy = proxy
            https_proxy = proxy

            proxies = {
                "http": http_proxy,
                "https": https_proxy,
            }

        return proxies

    def tornado_async_get(self, url, handler, method='GET', **kw):
        #if self.proxies:
            #kw['proxies'] = self.proxies

        kw['proxy_host'] = self.config.HTTP_PROXY_HOST
        kw['proxy_port'] = self.config.HTTP_PROXY_PORT

        logging.debug('Enqueueing %s for %s...' % (method, url))
        self.otto.enqueue(url, handler, method, **kw)

    def async_get(self, url, handler, method='GET', **kw):
        if self.proxies:
            kw['proxies'] = self.proxies

        logging.debug('Enqueueing %s for %s...' % (method, url))
        self.otto.enqueue(url, handler, method, **kw)

    def get(self, url):
        url = join(self.config.HOLMES_API_URL.rstrip('/'), url.lstrip('/'))
        return requests.get(url)

    def post(self, url, data={}):
        url = join(self.config.HOLMES_API_URL.rstrip('/'), url.lstrip('/'))
        return requests.post(url, data=data)

    def get_description(self):
        return "%s%sholmes-worker%s (holmes-api v%s)" % (
            Fore.BLUE,
            Style.BRIGHT,
            Style.RESET_ALL,
            __version__
        )

    def get_config_class(self):
        return Config

    def stop_work(self):
        try:
            self.post('/worker/%s/dead' % self.uuid, data={'worker_uuid': self.uuid})
        except ConnectionError:
            pass
        self.working = False

    def do_work(self):
        if self._ping_api():
            err = None
            job = self._load_next_job()
            if job and self._start_job(job['url']):
                try:
                    self._start_reviewer(job=job)
                except InvalidReviewError:
                    err = str(sys.exc_info()[1])
                    logging.error("Fail to review %s: %s" % (job['url'], err))

                self._complete_job(error=err)
            elif job:
                logging.debug('Could not start job for url "%s". Maybe other worker doing it?' % job['url'])

    def _start_reviewer(self, job):
        if job:
            logging.debug('Starting Review for [%s]' % job['url'])
            reviewer = Reviewer(
                api_url=self.config.HOLMES_API_URL,
                page_uuid=job['page'],
                page_url=job['url'],
                config=self.config,
                validators=self.validators,
                facters=self.facters,
                async_get=self.tornado_async_get,
                wait=self.otto.wait,
                wait_timeout=0  # max time to wait for all requests to finish
            )

            reviewer.review()

    def _ping_api(self):
        try:
            self.post('/worker/%s/alive' % self.uuid)
            return True
        except ConnectionError:
            logging.fatal('Fail to ping API [%s]. Stopping Worker.' % self.config.HOLMES_API_URL)
            self.stop_work()
            return False

    def _load_next_job(self):
        try:
            response = self.get('/next')
            if response and response.text:
                return loads(response.text)
        except ConnectionError:
            logging.fatal('Fail to get next review from [%s]. Stopping Worker.' % self.config.HOLMES_API_URL)
            self.stop_work()

        return None

    def _start_job(self, url):
        if not url:
            return False

        try:
            response = self.post('/worker/%s/start' % self.uuid, data=url)
            if response.text == 'OK':
                return True

            if response.text == 'NOK':
                logging.debug('Failed to acquire lock on %s.' % url)
                return False
        except ConnectionError:
            err = sys.exc_info()
            logging.error('Fail to start review: %s.' % err[1])

        return False

    def _complete_job(self, error=None):
        try:
            url = '/worker/%s/complete' % self.uuid
            response = self.post(url, data=dumps({'error': error}))
            return ('OK' == response.text)
        except ConnectionError:
            logging.error('Fail to complete worker.')

        return False

    def _load_validators(self):
        return load_classes(default=self.config.VALIDATORS)

    def _load_facters(self):
        return load_classes(default=self.config.FACTERS)


def main():
    worker = HolmesWorker(sys.argv[1:])
    worker.run()

if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

import sys
from os.path import join
import urlparse
import inspect
import email.utils as eut
from datetime import datetime

import codecs
from box.util.rotunicode import RotUnicode
codecs.register(RotUnicode.search_function)

import requests
from requests.exceptions import ConnectionError
import lxml.html
import logging
from ujson import dumps

from holmes.config import Config
from holmes.facters import Facter
from holmes.validators.base import Validator


class InvalidReviewError(RuntimeError):
    pass


class ReviewDAO(object):
    def __init__(self, page_uuid, page_url, last_modified=None, expires=None):
        self.page_uuid = page_uuid
        self.page_url = page_url
        self.last_modified = last_modified
        self.expires = expires
        self.facts = {}
        self.violations = []
        self.data = {}
        self._current = None

    def add_fact(self, key, value):
        self.facts[key] = {
            'key': key,
            'value': value
        }

    def add_violation(self, key, value, points):
        self.violations.append({
            'key': key,
            'value': value,
            'points': points
        })

    def to_dict(self):
        return {
            'page_uuid': self.page_uuid,
            'page_url': self.page_url,
            'facts': self.facts.values(),
            'violations': self.violations,
            'lastModified': self.last_modified,
            'expires': self.expires
        }


class Reviewer(object):
    def __init__(self, api_url, page_uuid, page_url, config=None, validators=[], facters=[], async_get=None, wait=None, wait_timeout=None):
        self.api_url = api_url

        self.page_uuid = page_uuid
        self.page_url = page_url

        self.review_dao = ReviewDAO(self.page_uuid, self.page_url)

        assert isinstance(config, Config), 'config argument must be an instance of holmes.config.Config'
        self.config = config

        for facter in facters:
            message = 'All facters must subclass holmes.facters.Facter (Error: %s)' % facter.__class__.__name__
            assert inspect.isclass(facter), message
            assert issubclass(facter, Facter), message

        for validator in validators:
            message = 'All validators must subclass holmes.validators.base.Validator (Error: %s)' % validator.__class__.__name__
            assert inspect.isclass(validator), message
            assert issubclass(validator, Validator), message

        self.validators = validators
        self.facters = facters

        self.responses = {}
        self.raw_responses = {}
        self.status_codes = {}

        self.proxies = None
        if self.config.HTTP_PROXY_HOST is not None:
            proxy = "%s:%s" % (self.config.HTTP_PROXY_HOST, self.config.HTTP_PROXY_PORT)
            http_proxy = proxy
            https_proxy = proxy

            self.proxies = {
                "http": http_proxy,
                "https": https_proxy,
            }

        self.async_get_func = async_get
        self._wait_for_async_requests = wait
        self._wait_timeout = wait_timeout

    def _async_get(self, url, handler, method='GET', **kw):
        if self.async_get_func:
            self.async_get_func(url, handler, method, **kw)

    def _get(self, url):
        #if self.proxies:
            #logging.debug('Getting "%s" using proxy "%s"...' % (url, self.proxies.get('http', None)))
            #return requests.get(url, proxies=self.proxies)

        return requests.get(url)

    def _post(self, url, data):
        #if self.proxies:
            #logging.debug('Posting to "%s" using proxy "%s"...' % (url, self.proxies.get('http', None)))
            #return requests.post(url, data=data, proxies=self.proxies)

        return requests.post(url, data=data)

    def review(self):
        self.load_content(self.content_loaded)
        self.wait_for_async_requests()

    def load_content(self, callback):
        self._async_get(self.page_url, callback)

    def content_loaded(self, url, response):
        if response.status_code > 399:
            if response.text:
                text = response.text.decode('rotunicode')
            else:
                text = 'Empty response.text'

            msg = "Could not load '%s' (%s) - %s!" % (url,
                                                      response.status_code,
                                                      text)
            logging.error(msg)
            return

        logging.debug('Content for url %s loaded.' % url)

        last_modified = None

        if 'Last-Modified' in response.headers:
            last_modified = datetime(*eut.parsedate(response.headers['Last-Modified'])[:6])

        self.review_dao.last_modified = last_modified

        expires = None
        if 'Expires' in response.headers:
            expires = datetime(*eut.parsedate(response.headers['Expires'])[:6])

        self.review_dao.expires = expires

        self._current = response

        try:
            self._current.html = lxml.html.fromstring(response.text)
        except (lxml.etree.XMLSyntaxError, lxml.etree.ParserError):
            self._current.html = None

            # can't do this there, this is not a validator
            #self.add_violation(
                #key='invalid.content',
                #title='Invalid Content',
                #description='Fail to parse content from %s' % url,
                #points=1000)

        self.run_facters()
        self.wait_for_async_requests()

        self.run_validators()
        self.wait_for_async_requests()

        self.save_review()

    @property
    def current(self):
        return self._current

    @property
    def current_html(self):
        if not hasattr(self.current, 'html') or self.current.html is None:
            return lxml.html.HtmlElement()
        else:
            return self.current.html

    def run_facters(self):
        for facter in self.facters:
            logging.debug('---------- Started running facter %s ---------' % facter.__name__)
            facter_instance = facter(self)
            facter_instance.get_facts()

    def run_validators(self):
        for validator in self.validators:
            logging.debug('---------- Started running validator %s ---------' % validator.__name__)
            validator_instance = validator(self)
            validator_instance.validate()

    def get_url(self, url):
        return join(self.api_url.rstrip('/'), url.lstrip('/'))

    def enqueue(self, urls):
        if not urls:
            return

        if isinstance(urls, basestring):
            post_url = self.get_url('/page')
            data = dumps({
                'url': urls,
                'origin_uuid': str(self.page_uuid)
            })
            error_message = "Could not enqueue page '" + urls + "'! Status Code: %d, Error: %s"
        else:
            post_url = self.get_url('/pages')
            data = {
                "url": urls,
                'origin_uuid': str(self.page_uuid)
            }
            error_message = "Could not enqueue the " + str(len(urls)) + " pages sent! Status Code: %d, Error: %s"

        response = self._post(post_url, data=data)

        if response.status_code > 399:
            logging.error(error_message % (
                response.status_code,
                response.text
            ))

    def add_fact(self, key, value):
        self.review_dao.add_fact(key, value)

    def add_violation(self, key, value, points):
        self.review_dao.add_violation(key, value, points)

    def save_review(self):
        url = self.get_url('/page/%s/review/' % (self.page_uuid))

        try:
            data = dumps(self.review_dao.to_dict())
            response = self._post(url, data={
                'review': data
            })
        except ConnectionError:
            err = sys.exc_info()[1]
            logging.error("Could not save review! ConnectionError - %s (%s)" % (
                url,
                str(err)
            ))
            return

        if response.status_code > 399:
            logging.error("Could not save review! Status Code: %d, Error: %s" % (
                response.status_code,
                response.text
            ))

    def wait_for_async_requests(self):
        self._wait_for_async_requests(self._wait_timeout)

    def is_root(self):
        result = urlparse.urlparse(self.page_url)
        return '{0}://{1}'.format(result.scheme, result.netloc) == self.page_url.rstrip('/')



#!/usr/bin/python
# -*- coding: utf-8 -*-

import logging

from six.moves.urllib.parse import urlparse

EMPTY_DOMAIN_RESULT = ('', '')


def get_domain_from_url(url, default_scheme='http'):
    if not url:
        return EMPTY_DOMAIN_RESULT

    scheme = default_scheme
    result = urlparse(url)
    domain = result.netloc

    if (result.scheme and result.scheme in ['http', 'https']):
        scheme = result.scheme

    if (result.scheme and result.scheme not in ['http', 'https']) and not domain:
        domain = result.scheme
    else:
        if not domain:
            domain = result.path.split('/')[0]
            if not '.' in domain:
                return EMPTY_DOMAIN_RESULT

        if ':' in domain:
            domain = domain.split(':')[0]

    original_domain = domain.strip()
    domain = original_domain.replace('www.', '')

    return domain, '%s://%s/' % (scheme, original_domain)


def get_class(klass):
    module_name, class_name = klass.rsplit('.', 1)

    module = __import__(module_name)

    if '.' in module_name:
        module = reduce(getattr, module_name.split('.')[1:], module)

    return getattr(module, class_name)


def load_classes(classes=None, classes_to_load=None, default=None):
    if classes_to_load is None:
        classes_to_load = default

    if classes is None:
        classes = []

    for class_full_name in classes_to_load:
        if isinstance(class_full_name, (tuple, set, list)):
            load_classes(classes, class_full_name)
            continue

        try:
            klass = get_class(class_full_name)
            classes.append(klass)
        except ValueError:
            logging.warn('Invalid class name [%s]. Will be ignored.' % class_full_name)
        except AttributeError:
            logging.warn('Class [%s] not found. Will be ignored.' % class_full_name)
        except ImportError:
            logging.warn('Module [%s] not found. Will be ignored.' % class_full_name)

    return classes



#!/usr/bin/python
# -*- coding: utf-8 -*-

import logging
from uuid import uuid4

from tornado.websocket import WebSocketHandler


class EventBusHandler(WebSocketHandler):
    def keep_alive(self):
        return True

    def open(self):
        self.uuid = uuid4()
        logging.debug("WebSocket opened.")
        self.application.event_bus.subscribe('events', self.uuid, self.async_callback(self.on_event))

    def on_event(self, message):
        if self.ws_connection is not None:
            self.write_message(message)

    def on_close(self):
        if hasattr(self, 'uuid'):
            self.application.event_bus.unsubscribe('events', self.uuid)
        logging.debug("WebSocket closed.")



"""Change fact to contain less info

Revision ID: 256faac818c5
Revises: 39091f5a3e5f
Create Date: 2013-12-02 21:29:59.676698

"""

# revision identifiers, used by Alembic.
revision = '256faac818c5'
down_revision = '39091f5a3e5f'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.drop_column('facts', 'title')
    op.drop_column('facts', 'unit')


def downgrade():
    op.add_column('facts', sa.Column('title', sa.String(2000), nullable=False))
    op.add_column('facts', sa.Column('unit', sa.String(2000), nullable=False))



#!/usr/bin/python
# -*- coding: utf-8 -*-

import sqlalchemy as sa

from holmes.models import Base, JsonType


class Fact(Base):
    __tablename__ = "facts"

    id = sa.Column(sa.Integer, primary_key=True)
    # key = sa.Column('key', sa.String(2000), nullable=False)
    value = sa.Column('value', JsonType, nullable=False)
    #title = sa.Column('title', sa.String(2000), nullable=False)
    #unit = sa.Column('unit', sa.String(2000), nullable=False)

    review_id = sa.Column('review_id', sa.Integer, sa.ForeignKey('reviews.id'))
    # review comes from Review relationship

    key_id = sa.Column('key_id', sa.Integer, sa.ForeignKey('keys.id'))

    def to_dict(self, fact_definitions):
        definition = fact_definitions.get(self.key.name, {})
        return {
            'title': definition.get('title', 'unknown'),
            'key': self.key.name,
            'unit': definition.get('unit', 'value'),
            'value': definition.get('description', lambda value: value)(self.value)
        }

    def __str__(self):
        return '%s: %s' % (self.key.name, self.value)

    def __repr__(self):
        return str(self)



#!/usr/bin/python
# -*- coding: utf-8 -*-

import sqlalchemy as sa
from sqlalchemy.orm import relationship
from sqlalchemy import func

from holmes.models import Base


class Domain(Base):
    __tablename__ = "domains"

    id = sa.Column(sa.Integer, primary_key=True)
    url = sa.Column('url', sa.String(2000), nullable=False)
    url_hash = sa.Column('url_hash', sa.String(128), nullable=False)
    name = sa.Column('name', sa.String(2000), nullable=False)

    pages = relationship("Page", backref="domain")
    reviews = relationship("Review", backref="domain")

    def to_dict(self):
        return {
            "url": self.url,
            "name": self.name
        }

    @classmethod
    def get_pages_per_domain(cls, db):
        from holmes.models import Page
        return dict(db.query(Page.domain_id, sa.func.count(Page.id)).group_by(Page.domain_id).all())

    def get_page_count(self, db):
        from holmes.models import Page
        return db.query(func.count(Page.id)).filter(Page.domain_id == self.id).scalar()

    @classmethod
    def get_violations_per_domain(cls, db):
        from holmes.models import Review, Violation

        violations = db \
            .query(Review.domain_id, sa.func.count(Violation.id).label('count')) \
            .filter(Violation.review_id == Review.id) \
            .filter(Review.is_active == True) \
            .group_by(Review.domain_id) \
            .all()

        domains = {}
        for domain in violations:
            domains[domain.domain_id] = domain.count

        return domains

    def get_violation_data(self, db):
        from holmes.models import Review, Violation

        result = db.query(sa.func.count(Violation.id).label('count')) \
            .join(Review, Violation.review_id == Review.id) \
            .filter(Review.domain_id == self.id, Review.is_active == True) \
            .one()

        return (
            result.count
        )

    def get_violations_per_day(self, db):
        from holmes.models import Review, Violation  # Prevent circular dependency

        violations = db \
            .query(
                sa.func.year(Review.completed_date).label('year'),
                sa.func.month(Review.completed_date).label('month'),
                sa.func.day(Review.completed_date).label('day'),
                sa.func.count(Violation.id).label('violation_count'),
                sa.func.sum(Violation.points).label('violation_points')
            ).join(
                Domain, Domain.id == Review.domain_id
            ).join(
                Violation, Violation.review_id == Review.id
            ).filter(Review.is_complete == True).filter(Review.domain_id == self.id) \
            .group_by(
                sa.func.year(Review.completed_date),
                sa.func.month(Review.completed_date),
                sa.func.day(Review.completed_date),
            ) \
            .order_by(Review.completed_date) \
            .all()

        result = []

        for day in violations:
            dt = "%d-%d-%d" % (day.year, day.month, day.day)
            result.append({
                "completedAt": dt,
                "violation_count": int(day.violation_count),
                "violation_points": int(day.violation_points)
            })

        return result

    def get_active_reviews(self, db, current_page=1, page_size=10):
        from holmes.models import Page  # Prevent circular dependency

        lower_bound = (current_page - 1) * page_size
        upper_bound = lower_bound + page_size

        items = db \
            .query(
                Page.url, Page.uuid, Page.last_review_date,
                Page.last_review_uuid, Page.violations_count
            ) \
            .filter(Page.last_review_date != None) \
            .filter(Page.domain == self) \
            .order_by('violations_count desc')[lower_bound:upper_bound]

        return items

    @classmethod
    def get_domain_by_name(self, domain_name, db):
        return db.query(Domain).filter(Domain.name == domain_name).first()

    def get_active_review_count(self, db):
        from holmes.models import Review

        return db.query(func.count(Review.id)).filter(Review.is_active == True, Review.domain_id == self.id).scalar()



#!/usr/bin/python
# -*- coding: utf-8 -*-

from datetime import datetime
from uuid import UUID, uuid4

from tornado import gen
from ujson import loads, dumps

from holmes.models import Review, Page
from holmes.handlers import BaseHandler


class BaseReviewHandler(BaseHandler):
    def _parse_uuid(self, uuid):
        try:
            return UUID(uuid)
        except ValueError:
            return None


class ReviewHandler(BaseReviewHandler):
    def get(self, page_uuid, review_uuid):
        review = None
        if self._parse_uuid(review_uuid):
            review = Review.by_uuid(review_uuid, self.db)

        if not review:
            self.set_status(404, 'Review with uuid of %s not found!' % review_uuid)
            return

        result = review.to_dict(self.application.fact_definitions, self.application.violation_definitions)
        result.update({
            'violationPoints': review.get_violation_points(),
            'violationCount': review.violation_count,
        })

        self.write_json(result)

    @gen.coroutine
    def post(self, page_uuid, review_uuid=None):
        page = Page.by_uuid(page_uuid, self.db)

        review_data = loads(self.get_argument('review'))

        page.expires = review_data['expires']
        if page.expires is not None:
            page.expires = datetime.utcfromtimestamp(page.expires)

        page.last_modified = review_data['lastModified']
        if page.last_modified is not None:
            page.last_modified = datetime.utcfromtimestamp(page.last_modified)

        self.db.flush()

        review = Review(
            domain=page.domain,
            page=page,
            is_active=True,
            is_complete=False,
            completed_date=datetime.utcnow(),
            uuid=uuid4(),
        )

        self.db.add(review)
        self.db.flush()

        for fact in review_data['facts']:
            name = fact['key']
            key = self.application.fact_definitions[name]['key']
            review.add_fact(key, fact['value'])

        self.db.flush()

        for violation in review_data['violations']:
            name = violation['key']
            key = self.application.violation_definitions[name]['key']
            review.add_violation(key, violation['value'], violation['points'])

        self.db.flush()

        page.violations_count = len(review_data['violations'])
        self.db.flush()

        review.is_complete = True
        self.db.flush()

        if not page.last_review:
            yield self.cache.increment_active_review_count(page.domain)

            yield self.cache.increment_violations_count(
                page.domain,
                increment=page.violations_count
            )
        else:
            old_violations_count = len(page.last_review.violations)
            new_violations_count = len(review.violations)

            yield self.cache.increment_violations_count(
                page.domain,
                increment=new_violations_count - old_violations_count
            )

            page.last_review.is_active = False
            self.db.flush()

        page.last_review_uuid = review.uuid
        page.last_review = review
        self.db.flush()

        page.last_review_date = review.completed_date
        self.db.flush()

        self._remove_older_reviews_with_same_day(review)

        self.application.event_bus.publish(dumps({
            'type': 'new-review',
            'reviewId': str(review.uuid)
        }))

    def _remove_older_reviews_with_same_day(self, review):
        dt = datetime.now()
        dt = dt.replace(hour=0, minute=0, second=0, microsecond=0)

        for review in self.db.query(Review) \
                .filter(Review.page == review.page) \
                .filter(Review.uuid != review.uuid) \
                .filter(Review.created_date >= dt) \
                .all():

            for fact in review.facts:
                self.db.delete(fact)

            self.db.flush()

            for violation in review.violations:
                self.db.delete(violation)

            self.db.flush()

            self.db.delete(review)

            self.db.flush()


class LastReviewsHandler(BaseReviewHandler):
    def get(self):
        reviews = Review.get_last_reviews(self.db)

        reviews_json = []
        for review in reviews:
            review_dict = review.to_dict(self.application.fact_definitions, self.application.violation_definitions)
            data = {
                'violationCount': review.violation_count,
            }
            review_dict.update(data)
            reviews_json.append(review_dict)

        self.write_json(reviews_json)



#!/usr/bin/python
# -*- coding: utf-8 -*-

import logging

from ujson import dumps
import tornado.web

from holmes.models.worker import Worker
from holmes.handlers import BaseHandler
from holmes.redis_lock import RedisLock


class WorkerStateHandler(BaseHandler):

    @tornado.web.asynchronous
    def post(self, worker_uuid, state):
        worker = Worker.by_uuid(worker_uuid, self.db)

        if not worker:
            self.set_status(404, 'Unknown Worker')
            self.finish()
            return

        if 'start' == state:
            url = self.request.body
            if not url:
                logging.warning('Invalid URL trying to start worker %s.' % worker_uuid)
                raise tornado.web.HTTPError(400, 'Invalid URL', reason='Invalid URL')

            self.start_work(worker, url)
            return

        elif 'complete' == state and worker.current_url is not None:
            self.complete_work(worker, worker.current_url)
            worker.current_url = None

        else:
            logging.warning('Invalid operation (not start nor complete) in worker %s.' % worker_uuid)
            raise tornado.web.HTTPError(400, 'Invalid Operation', reason='Invalid Operation')

    def start_work(self, worker, url):
        lock = RedisLock(
            self.application.redis,
            lock_key="redis-lock-%s" % url,
            lock_timeout=self.application.config.ZOMBIE_WORKER_TIME)

        lock.acquire(self.handle_lock_acquired(worker, url))

    def handle_lock_acquired(self, worker, url):
        def handle(acquired):
            if acquired:
                worker.current_url = url
                self.db.flush()

                self.application.event_bus.publish(dumps({
                    'type': 'worker-status',
                    'workerId': str(worker.uuid)
                }))

                self.write('OK')
                self.finish()
            else:
                self.write("NOK")
                self.finish()

        return handle

    def complete_work(self, worker, url):
        lock = RedisLock(
            self.application.redis,
            lock_key="redis-lock-%s" % url,
            lock_timeout=self.application.config.ZOMBIE_WORKER_TIME)

        lock.release(self.handle_lock_released(worker, url))

    def handle_lock_released(self, worker, url):
        def handle(deleted_count):
            worker.current_url = None
            self.db.flush()

            self.application.event_bus.publish(dumps({
                'type': 'worker-status',
                'workerId': str(worker.uuid)
            }))

            self.write('OK')
            self.finish()

        return handle



#!/usr/bin/python
# -*- coding: utf-8 -*-

from holmes.validators.base import Validator


class LinkWithRedirectValidator(Validator):
    @classmethod
    def get_link_with_redirect_message(cls, value):
        return 'Link with redirect, in most cases, should ' \
               'not be used. Redirects were found for ' \
               'link: %s.' % value

    @classmethod
    def get_violation_definitions(cls):
        return {
            'link.redirect.302': {
                'title': 'Link with 302 redirect',
                'description': cls.get_link_with_redirect_message
            },
            'link.redirect.307': {
                'title': 'Link with 307 redirect',
                'description': cls.get_link_with_redirect_message
            },
        }

    def validate(self):
        links = self.get_links()

        for url, response in links:
            if response.status_code in [302, 307]:
                self.add_violation(
                    key='link.redirect.%d' % response.status_code,
                    value=response.status_code,
                    points=10
                )

    def get_links(self):
        return self.review.data.get('page.links', None)



#!/usr/bin/python
# -*- coding: utf-8 -*-

import sqlalchemy as sa
from sqlalchemy.orm import relationship

from holmes.models import Base


class Key(Base):
    __tablename__ = "keys"

    id = sa.Column(sa.Integer, primary_key=True)
    name = sa.Column('name', sa.String(2000), nullable=False)

    facts = relationship("Fact", cascade="all,delete", backref="key")
    violations = relationship("Violation", cascade="all,delete", backref="key")

    def __str__(self):
        return '%s' % (self.name)

    def __repr__(self):
        return str(self)

    @classmethod
    def get_or_create(cls, db, key_name):
        key = db.query(Key).filter(Key.name == key_name).scalar()

        if not key:
            key = Key(name=key_name)

        return key



#!/usr/bin/python
# -*- coding: utf-8 -*-

from urlparse import urlparse

from holmes.validators.base import Validator


class LinkWithRelCanonicalValidator(Validator):

    @classmethod
    def get_absent_meta_canonical_message(cls, value):
        url = 'https://support.google.com/webmasters/answer/139394?hl=en'
        return 'As can be seen in this page <a href="%s">About ' \
               'rel="canonical"</a>, it\'s a good practice to ' \
               'include rel="canonical" urls in the pages for ' \
               'your website.' % url

    @classmethod
    def get_violation_definitions(cls):
        return {
            'absent.meta.canonical': {
                'title': 'Link with rel="canonical" not found',
                'description': cls.get_absent_meta_canonical_message
            }
        }

    def validate(self):
        if not self.config.FORCE_CANONICAL:
            # Only pages with query string parameters
            if self.page_url:
                if not self.is_valid(self.page_url):
                    return

                if not urlparse(self.page_url).query:
                    return

        head = self.get_head()
        if head:
            canonical = [item for item in head if item.get('rel') == 'canonical']

            if not canonical:
                self.add_violation(
                    key='absent.meta.canonical',
                    value=None,
                    points=30
                )

    def get_head(self):
        return self.review.data.get('page.head', None)



#!/usr/bin/python
# -*- coding: utf-8 -*-

from holmes.validators.base import Validator


class ImageRequestsValidator(Validator):

    @classmethod
    def get_broken_images_message(cls, value):
        return 'The image(s) in "%s" could not be found or took ' \
               'more than 10 seconds to load.' % (
                   ', '.join(
                   [
                       '<a href="%s" target="_blank">Link #%s</a>' % (url,
                                                                      index)
                       for index, url in enumerate(value)
                   ]))

    @classmethod
    def get_requests_images_message(cls, value):
        return 'This page has %d image requests (%d over limit). ' \
               'Having too many requests impose a tax in the browser ' \
               'due to handshakes.' % (value['total'], value['limit'])

    @classmethod
    def get_total_size_message(cls, value):
        return 'There`s %.2fkb of images in this page and that ' \
               'adds up to more download time slowing down the page ' \
               'rendering to the user.' % value

    @classmethod
    def get_single_image_size_message(cls, value):
        if 'over_max_size' in value:
            data = []
            for url, size_img in value['over_max_size']:
                size = size_img - value['limit']
                name = url.rsplit('/', 1)[-1]
                if size > 0:
                    desc = ' (%dkb above limit)' % (size)
                else:
                    desc = ''
                data.append(
                    '<a href="%s" target="_blank">%s</a>%s' % (url, name, desc)
                )

            return 'Some images are above the expected limit (%dkb): %s' % (
                value['limit'], ', '.join(data))

    @classmethod
    def get_violation_definitions(cls):
        return {
            'broken.img': {
                'title': 'Image not found.',
                'description': cls.get_broken_images_message
            },
            'single.size.img': {
                'title': 'Single image size in kb is too big.',
                'description': cls.get_single_image_size_message
            },
            'total.requests.img': {
                'title': 'Too many image requests.',
                'description': cls.get_requests_images_message
            },
            'total.size.img': {
                'title': 'Total image size in kb is too big.',
                'description': cls.get_total_size_message
            }
        }

    def validate(self):
        img_files = self.get_images()
        total_size = self.get_images_size()

        broken_imgs = set()

        over_max_size = set()

        for url, response in img_files:

            if response.status_code > 399:
                broken_imgs.add(url)

            if response.text is not None:
                size_img = len(response.text) / 1024.0

                if size_img > self.reviewer.config.MAX_KB_SINGLE_IMAGE:
                    over_max_size.add((url, size_img))

        if broken_imgs:
            self.add_violation(
                key='broken.img',
                value=broken_imgs,
                points=50
            )

        if over_max_size:
            points = 0
            for url, size_img in over_max_size:
                size = size_img - self.reviewer.config.MAX_KB_SINGLE_IMAGE
                points += size

            self.add_violation(
                key='single.size.img',
                value={
                    'limit': self.reviewer.config.MAX_KB_SINGLE_IMAGE,
                    'over_max_size': over_max_size
                },
                points=points
            )

        if len(img_files) > self.reviewer.config.MAX_IMG_REQUESTS_PER_PAGE:
            self.add_violation(
                key='total.requests.img',
                value={
                    'total': len(img_files),
                    'limit': len(img_files) - self.reviewer.config.MAX_IMG_REQUESTS_PER_PAGE
                },
                points=5 * (len(img_files) - self.reviewer.config.MAX_IMG_REQUESTS_PER_PAGE)
            )

        if total_size > self.reviewer.config.MAX_IMG_KB_PER_PAGE:
            self.add_violation(
                key='total.size.img',
                value=total_size,
                points=int(total_size - self.reviewer.config.MAX_IMG_KB_PER_PAGE)
            )

    def get_images(self):
        return self.review.data.get('page.images', None)

    def get_images_size(self):
        return self.review.data.get('total.size.img', 0)



#!/usr/bin/python
# -*- coding: utf-8 -*-

from holmes.validators.base import Validator


class TitleValidator(Validator):
    def required_facters(self):
        return ("holmes.facters.title.Title", )

    @classmethod
    def get_violation_definitions(cls):
        return {
            'page.title.not_found': {
                'title': 'Page title not found.',
                'description': lambda value: "Title was not found on '%s'." % value
            },
            'page.title.multiple': {
                'title': 'Too many titles.',
                'description': lambda value: "Page '%s' has %d title tags." % (value['page_url'], value['title_count'])
            },
            'page.title.size': {
                'title': 'Maximum size of a page title',
                'description': lambda value: 'Title is too long on "%s". The max size is %d characters.' % (value['page_url'], value['max_size'])
            }
        }

    def validate(self):
        title_count = self.review.data.get('page.title_count', 0)
        title = self.review.facts.get('page.title', None)

        if not title_count or not title:
            self.add_violation(
                key='page.title.not_found',
                value=self.reviewer.page_url,
                points=50
            )
            return

        if title_count > 1:
            self.add_violation(
                key='page.title.multiple',
                value={
                    'page_url': self.reviewer.page_url,
                    'title_count': title_count
                },
                points=50
            )
        elif len(title) > self.reviewer.config.MAX_TITLE_SIZE:
            self.add_violation(
                key='page.title.size',
                value={
                    'page_url': self.reviewer.page_url,
                    'max_size': self.reviewer.config.MAX_TITLE_SIZE
                },
                points=10
            )



#!/usr/bin/python
# -*- coding: utf-8 -*-

from uuid import UUID
import hashlib

from ujson import loads, dumps
from tornado import gen
import tornado.httpclient
import logging
from sqlalchemy import or_

from holmes.models import Page, Domain, Review
from holmes.utils import get_domain_from_url
from holmes.handlers import BaseHandler


class PageHandler(BaseHandler):

    @gen.coroutine
    def post(self):
        post_data = loads(self.request.body)
        url = post_data['url']

        domain_name, domain_url = get_domain_from_url(url)
        if not domain_name:
            self.set_status(400, 'Invalid url [%s]' % url)
            self.write_json({
                'reason': 'invalid_url',
                'url': url
            })
            return

        phost = self.application.config.HTTP_PROXY_HOST
        pport = self.application.config.HTTP_PROXY_PORT

        request = tornado.httpclient.HTTPRequest(
            url=url,
            proxy_host=phost,
            proxy_port=pport
        )

        logging.info('Obtaining "%s" using proxy "%s:%s"...' % (url, phost, pport))

        response = yield tornado.gen.Task(self.application.http_client.fetch, request)

        if response.code > 399:
            self.set_status(400, 'Invalid URL [%s]' % url)
            self.write_json({
                'reason': 'invalid_url',
                'url': url
            })
            return

        if response.effective_url != url:
            self.set_status(400, 'Redirect URL [%s]' % url)
            self.write_json({
                'reason': 'redirect',
                'url': url,
                'effectiveUrl': response.effective_url
            })
            return

        domains = self.db.query(Domain).filter(or_(
            Domain.name == domain_name,
            Domain.name == domain_name.rstrip('/'),
            Domain.name == "%s/" % domain_name
        )).all()

        if not domains:
            domain = None
        else:
            domain = domains[0]

        if not domain:
            url_hash = hashlib.sha512(domain_url).hexdigest()
            domain = Domain(url=domain_url, url_hash=url_hash, name=domain_name)
            self.db.add(domain)
            self.db.flush()

            self.application.event_bus.publish(dumps({
                'type': 'new-domain',
                'domainUrl': str(domain_url)
            }))

        pages = self.db.query(Page).filter(or_(
            Page.url == url,
            Page.url == url.rstrip('/'),
            Page.url == "%s/" % url
        )).all()

        if pages:
            page = pages[0]
        else:
            page = None

        if page:
            self.write(str(page.uuid))
            return

        url_hash = hashlib.sha512(url).hexdigest()

        page = Page(url=url, url_hash=url_hash, domain=domain)
        self.db.add(page)
        self.db.flush()

        yield self.cache.increment_page_count(domain)

        self.application.event_bus.publish(dumps({
            'type': 'new-page',
            'pageUrl': str(url)
        }))

        self.write(str(page.uuid))

    def get(self, uuid=''):
        uuid = UUID(uuid)

        page = Page.by_uuid(uuid, self.db)

        if not page:
            self.set_status(404, 'Page UUID [%s] not found' % uuid)
            return

        page_json = {
            "uuid": str(page.uuid),
            "url": page.url
        }

        self.write(page_json)


class PageReviewsHandler(BaseHandler):

    def get(self, uuid='', limit=10):
        uuid = UUID(uuid)

        page = Page.by_uuid(uuid, self.db)

        if not page:
            self.set_status(404, 'Page UUID [%s] not found' % uuid)
            return

        reviews = self.db.query(Review) \
            .filter(Review.page == page) \
            .filter(Review.is_complete == True) \
            .order_by(Review.completed_date.desc())[:limit]

        result = []
        for review in reviews:
            result.append({
                'uuid': str(review.uuid),
                'completedAt': review.completed_date,
                'violationCount': review.violation_count
            })

        self.write_json(result)


class PagesHandler(BaseHandler):

    @gen.coroutine
    def post(self):
        urls = self.get_arguments('url')

        if not urls:
            self.set_status(200)
            self.write('0')
            return

        all_domains = []
        for url in urls:
            domain_name, domain_url = get_domain_from_url(url.strip())
            if not domain_name:
                self.set_status(400, 'In the urls you posted there is an invalid URL: %s' % url)
                return
            all_domains.append((domain_name, domain_url))

        domains = self.add_domains(all_domains)

        existing_pages = self.db.query(Page).filter(Page.url.in_(urls)).all()
        if not existing_pages:
            existing_pages = []
        existing_pages_dict = dict([(page.url, page) for page in existing_pages])

        added_pages = []
        for url in set(urls):
            if url in existing_pages_dict:
                continue

            page_url = str(url)

            url_hash = hashlib.sha512(page_url).hexdigest()

            has_lock = yield self.cache.has_lock(url_hash)
            if has_lock:
                logging.debug('Lock found in page %s. Skipping...' % page_url)
                continue

            yield self.cache.lock_page(url_hash)

            domain_name, domain_url = get_domain_from_url(url.strip())

            logging.debug("Adding URL: %s" % url)

            domain = domains[domain_url]
            page = Page(url=page_url, url_hash=url_hash, domain=domain)
            self.db.add(page)
            self.db.flush()

            yield self.cache.increment_page_count(domain)
            added_pages.append(page)

        if added_pages:
            self.application.event_bus.publish(dumps({
                'type': 'new-page',
                'pageUrl': added_pages[0].url
            }))

        pages = list(added_pages)

        self.write(str(len(pages)))

    def add_domains(self, domains):
        resulting_domains = {}

        #existing_domains = self.db.query(Domain.id, Domain.url).filter(
            #or_(
                #Domain.url.in_([
                    #domain[1] for domain in domains
                #]),
                #Domain.url.in_([
                    #domain[1].rstrip('/') for domain in domains
                #]),
                #Domain.url.in_([
                    #"%s/" % domain[1] for domain in domains
                #])
            #)
        #).all()
        #existing_domain_dict = dict([(domain[0], domain[1]) for domain in existing_domains])

        for domain_name, domain_url in domains:
            if domain_url in resulting_domains:
                continue

            domain = self.db.query(Domain).filter(or_(
                Domain.url == domain_url,
                Domain.url == domain_url.rstrip('/')
            )).first()

            if domain:
                resulting_domains[domain_url] = domain
                continue

            key = domain_url.rstrip('/')
            domain = self.db.query(Domain).filter(Domain.url == key).first()
            if domain:
                resulting_domains[domain_url] = domain
                continue

            key = ("%s/" % domain_url)
            domain = self.db.query(Domain).filter(Domain.url == key).first()
            if domain:
                resulting_domains[domain_url] = domain
                continue

            resulting_domains[domain_url] = Domain(
                name=domain_name,
                url=domain_url,
                url_hash=hashlib.sha512(domain_url).hexdigest())

            self.db.add(resulting_domains[domain_url])

            self.application.event_bus.publish(dumps({
                'type': 'new-domain',
                'domainUrl': domain_url
            }))

        self.db.flush()
        return resulting_domains


class PageViolationsPerDayHandler(BaseHandler):

    def get(self, uuid):
        page = Page.by_uuid(uuid, self.db)

        if not page:
            self.set_status(404, 'Page UUID [%s] not found' % uuid)
            return

        violations_per_day = page.get_violations_per_day(self.db)

        page_json = {
            "violations": violations_per_day
        }

        self.write_json(page_json)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
from holmes import __version__

tests_require = [
    'mock',
    'nose',
    'coverage',
    'yanc',
    'preggy',
    'tox',
    'ipdb',
    'coveralls',
    'factory_boy',
    'sqltap',
    'sphinx',
]

setup(
    name='holmes',
    version=__version__,
    description='Holmes is a service to investigate your website health.',
    long_description='''
Holmes is a service to investigate your website health.
''',
    keywords='seo health web',
    author='Globo.com',
    author_email='appdev@corp.globo.com',
    url='',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: MacOS',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
    ],
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'cow-framework',
        'ujson',
        'requests',
        'lxml',
        'cssselect',
        'sheep==0.3.8',
        'pycurl',
        'alembic',
        'mysql-python',
        'six',
        'octopus-http',
        'toredis',
        'raven',
        'rotunicode'
    ],
    extras_require={
        'tests': tests_require,
    },
    entry_points={
        'console_scripts': [
            'holmes-api=holmes.server:main',
            'holmes-worker=holmes.worker:main',
        ],
    },
)



#!/usr/bin/python
# -*- coding: utf-8 -*-

from uuid import uuid4
from datetime import datetime

import sqlalchemy as sa
from sqlalchemy.orm import relationship

from holmes.models import Base


class Review(Base):
    __tablename__ = "reviews"

    id = sa.Column(sa.Integer, primary_key=True)
    is_active = sa.Column('is_active', sa.Boolean, default=False, nullable=False)
    is_complete = sa.Column('is_complete', sa.Boolean, default=False, nullable=False)
    uuid = sa.Column('uuid', sa.String(36), default=uuid4, nullable=False)

    created_date = sa.Column('created_date', sa.DateTime, default=datetime.utcnow, nullable=False)
    completed_date = sa.Column('completed_date', sa.DateTime, nullable=True)
    completed_day = sa.Column('completed_day', sa.Date, nullable=True)

    failure_message = sa.Column('failure_message', sa.String(2000), nullable=True)

    domain_id = sa.Column('domain_id', sa.Integer, sa.ForeignKey('domains.id'))
    page_id = sa.Column('page_id', sa.Integer, sa.ForeignKey('pages.id'))

    facts = relationship("Fact", cascade="all,delete", backref="review")
    violations = relationship("Violation", cascade="all,delete", backref="review")

    def to_dict(self, fact_definitions, violation_definitions):
        return {
            'page': self.page and self.page.to_dict() or None,
            'domain': self.domain and self.domain.name or None,
            'isComplete': self.is_complete,
            'uuid': str(self.uuid),
            'createdAt': self.created_date,
            'completedAt': self.completed_date,
            'facts': [fact.to_dict(fact_definitions) for fact in self.facts],
            'violations': [violation.to_dict(violation_definitions) for violation in self.violations]
        }

    def __str__(self):
        return str(self.uuid)

    def __repr__(self):
        return str(self)

    def add_fact(self, key, value):
        if self.is_complete:
            raise ValueError("Can't add anything to a completed review.")

        from holmes.models.fact import Fact  # to avoid circular dependency

        fact = Fact(key=key, value=value)

        self.facts.append(fact)

    def add_violation(self, key, value, points):
        if self.is_complete:
            raise ValueError("Can't add anything to a completed review.")

        from holmes.models.violation import Violation  # to avoid circular dependency

        violation = Violation(key=key, value=value, points=int(float(points)))

        self.violations.append(violation)

    @property
    def failed(self):
        return self.failure_message is not None

    @classmethod
    def get_last_reviews(cls, db, limit=12):
        return db.query(Review).filter(Review.is_active == True) \
                               .order_by(Review.completed_date.desc())[:limit]

    def get_violation_points(self):
        points = 0
        for violation in self.violations:
            points += violation.points
        return points

    @classmethod
    def by_uuid(cls, uuid, db):
        return db.query(Review).filter(Review.uuid == uuid).first()

    @property
    def violation_count(self):
        return len(self.violations)



#!/usr/bin/python
# -*- coding: utf-8 -*-

from tornado.gen import coroutine

from holmes.models import Domain
from holmes.handlers import BaseHandler


class DomainsHandler(BaseHandler):

    @coroutine
    def get(self):
        domains = self.db.query(Domain).order_by(Domain.name.asc()).all()

        if not domains:
            self.write("[]")
            return

        result = []

        for domain in domains:
            page_count = yield self.cache.get_page_count(domain)
            review_count = yield self.cache.get_active_review_count(domain)
            violation_count = yield self.cache.get_violation_count(domain)

            if page_count > 0:
                review_percentage = round(float(review_count) / page_count * 100, 2)
            else:
                review_percentage = 0

            result.append({
                "url": domain.url,
                "name": domain.name,
                "violationCount": violation_count,
                "pageCount": page_count,
                "reviewCount": review_count,
                "reviewPercentage": review_percentage
            })

        self.write_json(result)


class DomainDetailsHandler(BaseHandler):

    @coroutine
    def get(self, domain_name):
        domain = Domain.get_domain_by_name(domain_name, self.db)

        if not domain:
            self.set_status(404, 'Domain %s not found' % domain_name)
            return

        page_count = yield self.cache.get_page_count(domain)
        review_count = yield self.cache.get_active_review_count(domain)
        violation_count = yield self.cache.get_violation_count(domain)

        if page_count > 0:
            review_percentage = round(float(review_count) / page_count * 100, 2)
        else:
            review_percentage = 0

        domain_json = {
            "name": domain.name,
            "url": domain.url,
            "pageCount": page_count,
            "reviewCount": review_count,
            "violationCount": violation_count,
            "reviewPercentage": review_percentage
        }

        self.write_json(domain_json)


class DomainViolationsPerDayHandler(BaseHandler):

    def get(self, domain_name):
        domain = Domain.get_domain_by_name(domain_name, self.db)

        if not domain:
            self.set_status(404, 'Domain %s not found' % domain_name)
            return

        violations_per_day = domain.get_violations_per_day(self.db)

        domain_json = {
            "name": domain.name,
            "url": domain.url,
            "violations": violations_per_day
        }

        self.write_json(domain_json)


class DomainReviewsHandler(BaseHandler):

    def get(self, domain_name):
        current_page = int(self.get_argument('current_page', 1))
        page_size = 10

        domain = Domain.get_domain_by_name(domain_name, self.db)

        if not domain:
            self.set_status(404, 'Domain %s not found' % domain_name)
            return

        reviews = domain.get_active_reviews(
            self.db,
            current_page=current_page,
            page_size=page_size
        )

        result = {
            'domainName': domain.name,
            'domainURL': domain.url,
            'pages': [],
        }

        for page in reviews:
            result['pages'].append({
                "url": page.url,
                "uuid": str(page.uuid),
                "violationCount": page.violations_count,
                "completedAt": page.last_review_date,
                "reviewId": str(page.last_review_uuid)
            })

        self.write_json(result)



#!/usr/bin/python
# -*- coding: utf-8 -*-

import logging
import time


class RedisLock(object):
    def __init__(self, redis, lock_key='redis-lock', lock_timeout=10):
        self.redis = redis
        self.lock_key = lock_key
        self.lock_timeout = lock_timeout
        self.callback = None
        self.release_callback = None

    def exec_callback(self, *args, **kw):
        if self.callback is not None:
            self.callback(*args, **kw)

    def acquire(self, callback):
        self.callback = callback
        self.now = int(time.time())
        self.redis.setnx(self.lock_key, self.now + self.lock_timeout + 1, self.handle_setnx)

    def handle_setnx(self, result):
        if result == 1 or result is True:
            self.acquire_time = self.now
            logging.debug('Lock key %s acquired (with setnx).' % self.lock_key)
            self.exec_callback(True)
            return

        self.redis.get(self.lock_key, self.handle_get_current_lock)

    def handle_get_current_lock(self, current_lock_timestamp):
        if not current_lock_timestamp:
            logging.debug('Lock key %s could not be acquired (by getting current lock).' % self.lock_key)
            self.exec_callback(False)
            return

        self.current_lock_timestamp = int(current_lock_timestamp)

        if self.now > self.current_lock_timestamp:
            self.redis.getset(self.lock_key, self.now + self.lock_timeout + 1, self.handle_next_lock_timestamp)
        else:
            logging.debug('Lock key %s could not be acquired (because lock timestamp is greater than now).' % self.lock_key)
            self.exec_callback(False)

    def handle_next_lock_timestamp(self, next_lock_timestamp):
        if not next_lock_timestamp:
            logging.debug('Lock key %s could not be acquired (due to next lock timestamp).' % self.lock_key)
            self.exec_callback(False)
            return

        next_lock_timestamp = int(next_lock_timestamp)

        if next_lock_timestamp == self.current_lock_timestamp:
            self.acquire_time = self.now
            logging.debug('Lock key %s acquired.' % self.lock_key)
            self.exec_callback(True)
            return

        logging.debug('Lock key %s could not be acquired (invalid timestamp).' % self.lock_key)
        self.exec_callback(False)

    def release(self, callback):
        self.release_callback = callback
        self.redis.delete(self.lock_key, callback=self.handle_lock_key_deleted)

    def handle_lock_key_deleted(self, deleted_count):
        logging.debug('Lock key %s deleted.' % self.lock_key)
        self.release_callback(deleted_count)



#!/usr/bin/python
# -*- coding: utf-8 -*-

from holmes.validators.base import Validator


class JSRequestsValidator(Validator):
    @classmethod
    def get_requests_js_message(cls, value):
        return 'This page has %d JavaScript request ' \
               '(%d over limit). Having too many requests impose ' \
               'a tax in the browser due to handshakes.' % (
                   value['total_js_files'],
                   value['over_limit'])

    @classmethod
    def get_total_size_message(cls, value):
        return 'There\'s %.2fkb of JavaScript in this page and ' \
               'that adds up to more download time slowing down ' \
               'the page rendering to the user.' % value

    @classmethod
    def get_violation_definitions(cls):
        return {
            'total.requests.js': {
                'title': 'Too many javascript requests.',
                'description': cls.get_requests_js_message
            },
            'total.size.js': {
                'title': 'Javascript size in kb is too big.',
                'description': cls.get_total_size_message
            }
        }

    def validate(self):
        total_js_files = self.get_total_requests_js()
        total_size_gzip = self.get_total_size_js_gzipped()

        max_requests = self.reviewer.config.MAX_JS_REQUESTS_PER_PAGE
        over_limit = total_js_files - max_requests

        if total_js_files > max_requests:
            self.add_violation(
                key='total.requests.js',
                value={
                    'total_js_files': total_js_files,
                    'over_limit': over_limit},
                points=5 * over_limit
            )

        max_kb_gzip = self.reviewer.config.MAX_JS_KB_PER_PAGE_AFTER_GZIP

        if total_size_gzip > max_kb_gzip:
            self.add_violation(
                key='total.size.js',
                value=total_size_gzip,
                points=int(total_size_gzip - max_kb_gzip)
            )

    def get_js_requests(self):
        return self.review.data.get('page.js', None)

    def get_total_requests_js(self):
        return self.review.data.get('total.requests.js', 0)

    def get_total_size_js(self):
        return self.review.data.get('total.size.js', 0)

    def get_total_size_js_gzipped(self):
        return self.review.data.get('total.size.js.gzipped', 0)



#!/usr/bin/python
# -*- coding: utf-8 -*-

from holmes.validators.base import Validator


class MetaTagsValidator(Validator):

    @classmethod
    def get_violation_definitions(cls):
        return {
            'absent.metatags': {
                'title': 'Meta tags not present.',
                'description': lambda value: "No meta tags found on this page. This is damaging for Search Engines."
            }
        }

    def validate(self):
        meta_tags = self.get_meta_tags()

        if not meta_tags:
            self.add_violation(
                key='absent.metatags',
                value='No metatags.',
                points=100
            )

    def get_meta_tags(self):
        return self.review.data.get('meta.tags', None)



#!/usr/bin/python
# -*- coding: utf-8 -*-

from sqlalchemy import or_

from holmes.models import Page
from holmes.handlers import BaseHandler


class SearchHandler(BaseHandler):

    def get(self):
        term = self.get_argument('term')

        page = self.db.query(Page) \
            .filter(or_(
                Page.url == term,
                Page.url == term.rstrip('/')
            )) \
            .filter(Page.last_review != None) \
            .first()

        if page is None:
            self.write_json(None)
            return

        self.write_json({
            "uuid": str(page.uuid),
            "url": page.url,
            "reviewId": str(page.last_review.uuid)
        })
